;; Generated package description from nlinum.el  -*- no-byte-compile: t -*-
(define-package "nlinum" "1.9.0.20221221.80940" "Show line numbers in the margin" 'nil :commit "17d4e9cd12480df64e5a64f80ef807459b18b35b" :url "https://elpa.gnu.org/packages/nlinum.html" :authors '(("Stefan Monnier" . "monnier@iro.umontreal.ca")) :maintainer '("Stefan Monnier" . "monnier@iro.umontreal.ca") :keywords '("convenience"))
